/*
 * Copyright (c) 2015 Tricoire Sebastien 3dsman@free.fr
 *
 * This software is provided 'as-is', without any express or implied
 * warranty.  In no event will the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 * claim that you wrote the original software. If you use this software
 * in a product, an acknowledgment in the product documentation would be
 * appreciated but is not required.
 * 2. Altered source versions must be plainly marked as such, and must not be
 * misrepresented as being the original software.
 * 3. This notice may not be removed or altered from any source distribution.
 *
 */

#ifndef OE_UTILS_H
#define OE_UTILS_H

#include <math.h>
#include <vector>
#include "Archive.h"



struct vector_2d {
    float x, y;
    vector_2d( float _x,  float _y) {
        x = _x;
        y = _y;
    }
	vector_2d() {}
	void persist(Pakal::Archive* archive)
	{
		archive->value("x", x);
		archive->value("y", y);
	}
	
	 inline vector_2d operator *(const vector_2d& pt) const{
		return {pt.x*x,pt.y*y};
	}
	 inline vector_2d operator *( const float val) {
		return {val*x,val*y};
	}
	inline vector_2d operator /(const vector_2d& pt) const{
		return {x/pt.x,y/pt.y};
	}
	 inline vector_2d operator /( const float val) {
		return {x/val,y/val};
	}
	 inline vector_2d operator +(const vector_2d pt) const{
		return {pt.x+x,pt.y+y};
	}
	 inline vector_2d operator +(const float val) const{
		return {val+x,val+y};
	}
	 inline vector_2d operator -(const vector_2d pt) const{
		return {x-pt.x,y-pt.y};
	}
	 inline vector_2d operator -(const float val) const{
		return {x-val,y-val};
    }
	 inline vector_2d operator -() const{
		return {-x,-y};
	}
     double dot(const vector_2d pt) const{
        return {x*pt.x+y*pt.y};
    }
	
	/** \brief normalise the vector (his length become 1)
	 */
    void normalize() {
        float length = 1/len() ;
        x *= length;
        y *= length;
    }
	
		
	/** \brief turn the vector 90° to the right
	 */
    void turnRight() {
		float tmpx = x;
        x = -y;
        y = tmpx;
    }
	
    void turnLeft() {
		float tmpx = x;
        x = y;
        y = -tmpx;
    }
	
    float len() {
		return sqrt(x*x+y*y);
    }
};

class BoundingBox
{
    private:
        /** \brief The bounding box values. */
        vector_2d min;
        vector_2d max;
    public:
        bool init = false;

        BoundingBox(){};
        BoundingBox(float minX, float minY, float maxX, float maxY);
        ~BoundingBox(){};
        const vector_2d& getMin();
        const vector_2d& getMax();
		const vector_2d getCenter();
        BoundingBox operator + (const BoundingBox& box);
        BoundingBox operator + (const vector_2d& point);
        void operator += (const BoundingBox& box);
        void operator += (const vector_2d& point);
};

inline std::vector<vector_2d> normals(std::vector<vector_2d> segments)
{
			vector_2d tmpNormal;
			vector_2d normal;
			std::vector<vector_2d> out;
			for(unsigned i=0; i<segments.size()-1; i++)
			{
				normal = segments.at(i+1)-segments.at(i);
				normal.turnRight();
				if (i==0)
					tmpNormal = normal;
				tmpNormal =tmpNormal + normal;
				tmpNormal.normalize();
				out.push_back( tmpNormal);
			
				tmpNormal = normal;
			}
			tmpNormal.normalize();
			out.push_back( tmpNormal);
			return out;
}

inline std::vector<vector_2d> subd(vector_2d pt1, vector_2d pt2, float maxlen)
{
	std::vector<vector_2d> out;
	vector_2d tmpvec = pt2 - pt1;
	float len = tmpvec.len();
	if (!len) return out;
	if (len<maxlen)
	{
		out.push_back( pt2);
		return out;
	}
	
	int nbsub = ceilf(len / maxlen);
	len /= (float)nbsub;
	tmpvec.normalize();
	tmpvec = tmpvec*len;
	for (;nbsub>0;nbsub--)
	{
		pt1 = pt1 + tmpvec;
		out.push_back( pt1);
	}
	return out;
}


/* inline vector_2d operator *( const vector_2d vect, const float val)  {
	return {val*vect.x,val*vect.y};
}*/
 inline vector_2d operator *( const float val, const vector_2d vect)  {
	return {val*vect.x,val*vect.y};
}
 /*inline vector_2d operator -( const vector_2d vect1, const vector_2d vect2)  {
	return {vect1.x-vect2.x,vect1.y-vect2.y};
}*/
 inline vector_2d operator -( const float val, const vector_2d vect)  {
	return {val-vect.x,val-vect.y};
}
 inline vector_2d operator -( const vector_2d vect, const float val)  {
	return {vect.x-val , vect.y-val};
}

 struct vector_2dt {
     vector_2d v;
     float t;

     vector_2dt( vector_2d _v,  float _t) {
         v = _v;
         t = _t;
     }

     vector_2dt() {}

     operator vector_2d() { return v; }

     inline vector_2dt operator *( const float val) {
        return {v*val, t*val};
    }

     inline vector_2dt operator +(const vector_2dt pt) const{
        return {pt.v+v, pt.t+t};
    }

     inline vector_2dt operator -(const vector_2dt pt) const{
        return {v-pt.v, t-pt.t};
    }

 };

 inline std::vector<vector_2d> normals(std::vector<vector_2dt> segments)
 {
 vector_2d tmpNormal;
 vector_2d normal;
 std::vector<vector_2d> out;
 for(unsigned i=0; i<segments.size()-1; i++)
 {
     normal = segments.at(i+1).v-segments.at(i).v;
     normal.turnRight();
     if (i==0)
         tmpNormal = normal;
     tmpNormal =tmpNormal + normal;
     tmpNormal.normalize();
     out.push_back( tmpNormal);

     tmpNormal = normal;
 }
 tmpNormal.normalize();
 out.push_back( tmpNormal);
 return out;
 }

static inline float minf(float a, float b) { return a < b ? a : b; }
static inline float maxf(float a, float b) { return a > b ? a : b; }

struct OE_Matrix
{
	float a, b, c, d, e, f;
	OE_Matrix() : a(1.0), b(0), c(0), d(1.0), e(0), f(0) {}
	OE_Matrix(float a, float b, float c, float d, float e, float f) : a(a), b(b), c(c), d(d), e(e), f(f) {}
	OE_Matrix(float* arr) : a(arr[0]), b(arr[1]), c(arr[2]), d(arr[3]), e(arr[4]), f(arr[5]) {}
	OE_Matrix operator*(const OE_Matrix& m)
	{
		/*return OE_Matrix(a*m.a+b*m.c, c*m.a+d*m.c, e*m.a+f*m.c+m.e,
		                 b*m.b+a*m.d, d*m.b+c*m.d, f*m.b+e*m.d+m.f);
		return OE_Matrix(a*m.a+b*m.c, b*m.b+a*m.d, c*m.a+d*m.c, d*m.b+c*m.d, e*m.a+f*m.c+m.e, f*m.b+e*m.d+m.f);*/
		return OE_Matrix(a*m.a+b*c, a*m.b+m.d*b, c*m.a+c*d, m.b*c+m.d*d, e*m.a+c*f+m.e, m.b*e+m.d*f+m.f);
	}
	void apply(float x, float y, float* newx, float* newy)
	{
		if(newx && newy)
		{
			*newx = a*x+c*y+e;
			*newy = b*x+d*y+f;
		}
	}
	void apply(float& x, float& y) const
	{
		float oldx = x;
		float oldy = y;
		x = a*oldx+c*oldy+e;
		y = b*oldx+d*oldy+f;
	}
};
		
#endif // OE_UTILS_H
